/*
 * Argus Client Software.  Tools to read, analyze and manage Argus data.
 * Copyright (c) 2000-2002 QoSient, LLC
 * Copyright (c) 2003,2005 Olaf Gellert
 * All rights reserved.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/*
 * raprelude - log argus records to a prelude-ids management console
 *
 * written by Olaf Gellert (Presecure Consulting GmbH) on base of the
 * argus-clients sources from Carter Bullard (QoSient, LLC)
 *
 * 2005-09-13: first changes to comply to prelude 0.9
 *
 */


/* include the usual system stuff */
#include <time.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <netdb.h>

#include <signal.h>
#include <ctype.h>
#include <compat.h>
#include <syslog.h>

#include <stdio.h>
#include <string.h>
#include <sys/time.h>


#if defined(HAVE_SOLARIS)
#include <sys/types.h>
#include <sys/socket.h>
#endif

/* include all the argus-stuff */
#include <argus_client.h>

/* include all the stuff for the IDS connections of argus clients */
#include <libprelude/prelude.h>

/* include our own stuff */
#include <raprelude.h>

int RaInitialized = 0;

char *idmef_all;
char *idmef_classif_name, *idmef_classif_url;
char *idmef_classif_origin, *idmef_impact_description;
char *idmef_impact_severity, *idmef_impact_completion;
char *idmef_impact_type, *idmef_confidence_rating;

char class_file_name[LINE_LEN];
char *classfile = &class_file_name[0];
struct ClassStruct *mclass = NULL;

int idmef_classif_origin_i;
int idmef_impact_type_i, idmef_impact_severity_i;
int idmef_impact_completion_i, idmef_confidence_rating_i;

prelude_client_t *pclient;
idmef_analyzer_t *argus_analyzer = NULL;

#define CONFIG_WORDS_COUNT 10
char *config_words[CONFIG_WORDS_COUNT] = {
  "PROTO",
  "SADDR", "SPORT",
  "DADDR", "DPORT",
  "ITYPE",
  "CLASS", "URL", "SEVERITY",
  "ACTION",
  };

#define STATE_PROTO    0
#define STATE_SADDR    1
#define STATE_SPORT    2
#define STATE_DADDR    3 
#define STATE_DPORT    4
#define STATE_ITYPE    5
#define STATE_CLASS    6   
#define STATE_URL      7
#define STATE_SEVERITY 8
#define STATE_ACTION   9

char *parse_error[16] = {
  "error parsing classification line",
  "could not split line into words",
  "no protocol specified",
  "unknown protocol specified",
  "no SRC specified",
  "invalid source port specification",
  "invalid source port operator",
  "no DST specified",
  "invalid destination port specification",
  "invalid destination port operator",
  "invalid icmp type specification",
  "no classification string found",
  "invalid severity level specified",
  "invalid action specified",
  "unknown option in configuration",
  "port specified for protocol other than TCP or UDP",
  };

#define PARSERR_GENERAL  0
#define PARSERR_NOSPLIT  1
#define PARSERR_NOPROTO  2
#define PARSERR_INVPROTO 3
#define PARSERR_NOSRC    4
#define PARSERR_SPORT	 5
#define PARSERR_SPORTOP  6
#define PARSERR_NODST	 7
#define	PARSERR_DPORT	 8
#define PARSERR_DPORTOP  9
#define PARSERR_ITYPE   10
#define PARSERR_SEVER   11
#define PARSERR_NOCLASS	12
#define PARSERR_ACTION  13
#define PARSERR_INVWORD 14
#define	PARSERR_IPORT	15


int Parse_idmef_impact_severity(char *severity) {
  if (strncmp(severity, "low", strlen("low")+1)==0) {
    return(IDMEF_IMPACT_SEVERITY_LOW);
    }
  else if (strncmp(severity, "medium", strlen("medium")+1)==0) {
    return(IDMEF_IMPACT_SEVERITY_MEDIUM);
    }
  else if (strncmp(severity, "high", strlen("high")+1)==0) {
    return(IDMEF_IMPACT_SEVERITY_HIGH);
    }
  else if (strncmp(severity, "info", strlen("info")+1)==0) {
    return(IDMEF_IMPACT_SEVERITY_INFO);
    }

  return(IDMEF_IMPACT_SEVERITY_ERROR);
  }

int RaParseCompleting = 0;

void
RaParseComplete (int sig)
{
   if (!RaParseCompleting) {
      RaParseCompleting++;
   }
}

void
ArgusClientTimeout ()
{
}

void
parse_arg (int argc, char**argv)
{ 
}


void
usage ()
{
   extern char version[];
   fprintf (stderr, "raprelude Version %s\n", IDMEF_ANALYZER_VERSION);
   fprintf (stderr, "usage: %s \n", ArgusProgramName);
   fprintf (stderr, "usage: %s [options] -S remoteServer  [- filter-expression]\n", ArgusProgramName);
   fprintf (stderr, "usage: %s [options] -r argusDataFile [- filter-expression]\n", ArgusProgramName);
   fprintf (stderr, "options: -a                 print record summaries on termination.\n");
   fprintf (stderr, "         -A                 print application bytes.\n");
   fprintf (stderr, "         -b                 dump packet-matching code.\n");
   fprintf (stderr, "         -C                 treat the remote source as a Cisco Netflow source.\n");
   fprintf (stderr, "         -d <bytes>         print number of <bytes> from user data capture buffer.\n");
   fprintf (stderr, "                   format:  num | s<num> | d<num> | s<num>:d<num>\n");
#if defined (ARGUSDEBUG)
   fprintf (stderr, "         -D <level>         specify debug level\n");
#endif
   fprintf (stderr, "         -e <encode>        convert user data using <encode> method.\n");
   fprintf (stderr, "                            Supported types are <Ascii> and <Encode64>.\n");
   fprintf (stderr, "         -E <file>          write records that are rejected by the filter into <file>\n");
   fprintf (stderr, "         -F <conffile>      read configuration from <conffile>.\n");
   fprintf (stderr, "         -h                 print help.\n");
   fprintf (stderr, "         -n                 don't convert numbers to names.\n");
   fprintf (stderr, "         -p <digits>        print fractional time with <digits> precision.\n");
   fprintf (stderr, "         -q                 quiet mode. don't print record outputs.\n");
   fprintf (stderr, "         -r <file>          read argus data <file>. '-' denotes stdin.\n");
   fprintf (stderr, "         -R                 print out response data when availabile.\n");
   fprintf (stderr, "         -s [-][+[#]]field  specify fields to print.\n");
   fprintf (stderr, "         -S <host[:port]>   specify remote argus <host> and optional port number.\n");
   fprintf (stderr, "         -t <timerange>     specify <timerange> for reading records.\n");
   fprintf (stderr, "                   format:  timeSpecification[-timeSpecification]\n");
   fprintf (stderr, "                            timeSpecification: [[[yyyy/]mm/]dd.]hh[:mm[:ss]]\n");
   fprintf (stderr, "                                                 [yyyy/]mm/dd\n");
   fprintf (stderr, "                                                 -%%d{yMdhms}\n");
   fprintf (stderr, "         -T <secs>          attach to remote server for T seconds.\n");
   fprintf (stderr, "         -u                 print time in Unix time format.\n");
#ifdef ARGUS_SASL
   fprintf (stderr, "         -U <user/auth>     specify <user/auth> authentication information.\n");
#endif
   fprintf (stderr, "         -w <file>          write output to <file>. '-' denotes stdout.\n");
   fprintf (stderr, "         -z                 print Argus TCP state changes.\n");
   fprintf (stderr, "         -Z <s|d|b>         print actual TCP flag values.<'s'rc | 'd'st | 'b'oth>\n");
   exit(1);
}

int RaLabelCounter = 0;

int port_match(arg_uint16 port, arg_uint16 port_action,
               arg_uint16 port_low, arg_uint16 port_hi) {
  switch (port_action) {
    case PORT_EQ:
      return(port == port_low); break;
    case PORT_GT:
      return(port > port_low); break;
    case PORT_LT:
      return(port < port_low); break;
    case PORT_NE:
      return(port != port_low); break;
    case PORT_RANGE:
      return((port >= port_low) && (port <=port_hi));
    default:
      return 0;
    } // end of switch
  }

struct ClassStruct *find_class(struct ClassStruct *start, struct ArgusRecord *argus) {
  struct ClassStruct *act, *next;
  unsigned int i, aproto, asaddr, adaddr;
  arg_uint16 asport, adport, aitype;

  act=start;
  // ArgusLog(LOG_INFO, "Processing record: %s", get_argus_string(argus));
  while (act != NULL) {
    if ((argus->ahdr.status & 0xFFF) == ETHERTYPE_IP) {
      aproto = argus->argus_far.flow.ip_flow.ip_p;
      asaddr = argus->argus_far.flow.ip_flow.ip_src;
      adaddr = argus->argus_far.flow.ip_flow.ip_dst;
      if ((act->proto != PROTO_ANY) && (act->proto != aproto)) {
        // protocol does not match, so try next classification entry
        act=act->nxt; continue;
        }
      if (!((asaddr & ~act->src.mask) == act->src.addr)) {
        // src address does not match
        act=act->nxt; continue;
        }
      if (!((adaddr & ~act->dst.mask) == act->dst.addr)) {
        // src address does not match
        act=act->nxt; continue;
        }
      if ((aproto == IPPROTO_TCP) || (aproto == IPPROTO_UDP)) {
        // compare port numbers
        asport = argus->argus_far.flow.ip_flow.sport;
        adport = argus->argus_far.flow.ip_flow.dport;
        if (act->src_port_action != NO_PORT) {
          if (! (port_match(asport, act->src_port_action,
                            act->src_port_low, act->src_port_hi))) {
            act=act->nxt; continue;
            } }
        if (act->dst_port_action != NO_PORT) {
          if (! (port_match(adport, act->dst_port_action,
                            act->dst_port_low, act->dst_port_hi))) {
            act=act->nxt; continue;
            } }
        }
      else if (aproto == IPPROTO_ICMP) {
        // if port specified: this class rule is not for ICMP, so go on
        if ((act->src_port_action != NO_PORT) ||
            (act->dst_port_action != NO_PORT)) {
           act=act->nxt; continue;
           }
        // compare icmp type
        aitype = argus->argus_far.flow.icmp_flow.type;
        if ((act->itype != NO_ITYPE) && (act->itype != ITYPE_ANY) &&
            (act->itype != aitype)) {
          act=act->nxt; continue;
          }
        }
      }
    else {
      // not ETHERTYPE_IP (ARP, REVARP)
      }
    // all checks were true so this is a matching classification entry
    return(act);
    }

  // this should never be reached (because at least the last
  // classification entry (automatically added by raprelude)
  // should always match)
  ArgusLog(LOG_INFO, "no classification entry matched!");
  return(NULL);
  }

int add_int_additionalData (idmef_alert_t *alert, const char *meaning,
  unsigned int data) {
  
  int ret;
  prelude_string_t *str;
  idmef_additional_data_t *ad;
  
  ret = idmef_alert_new_additional_data(alert, &ad, -1);
  if ( ret < 0 ) {
          return ret;
          }

  idmef_additional_data_set_integer(ad, data);

  ret = idmef_additional_data_new_meaning(ad, &str);
  if ( ret < 0 ) {
	  ArgusLog(LOG_INFO, "error creating additional-data meaning: %s",
		prelude_strerror(ret));
    return ret;
	  }
  ret = prelude_string_set_ref(str, meaning);
  if ( ret < 0 ) {
          ArgusLog(LOG_INFO, "error setting integer data meaning: %s.\n",
                       prelude_strerror(ret));
          return -1;
  }
  
  return 0;
}

int add_string_additionalData (idmef_alert_t *alert, const char *meaning,
         const char *data) {
         
         int ret;
         prelude_string_t *str;
         idmef_additional_data_t *ad;
         size_t size;
         
        if ( ! data || ! size )
                return 0;
        
        ret = idmef_alert_new_additional_data(alert, &ad, -1);
        if ( ret < 0 )
                return ret;


        size=strlen(data);
        ret = idmef_additional_data_set_string_ref_fast(ad, data, size);
        if ( ret < 0 ) {
                ArgusLog(LOG_INFO, "error setting string data: %s.\n",
                             prelude_strerror(ret));
                return -1;
        }

        ret = idmef_additional_data_new_meaning(ad, &str);
        if ( ret < 0 ) {
                ArgusLog(LOG_INFO, "error creating additional-data meaning: %s.\n",
                             prelude_strerror(ret));
                return -1;
        }
        
        ret = prelude_string_set_ref(str, meaning);
        if ( ret < 0 ) {
                ArgusLog(LOG_INFO, "error setting byte string data meaning: %s.\n",
                             prelude_strerror(ret));
        }
                
        return -1;

  }

int add_bin_additionalData (idmef_alert_t *alert, const char *meaning,
	 const unsigned char *data, size_t size) {

          int ret;
        prelude_string_t *str;
        idmef_additional_data_t *ad;

        if ( ! data || ! size )
                return 0;
        
        ret = idmef_alert_new_additional_data(alert, &ad, -1);
        if ( ret < 0 )
                return ret;

        ret = idmef_additional_data_set_byte_string_ref(ad, data, size);
        if ( ret < 0 ) {
                ArgusLog(LOG_INFO, "error setting byte string data: %s.\n",
                             prelude_strerror(ret));
                return -1;
        }

        ret = idmef_additional_data_new_meaning(ad, &str);
        if ( ret < 0 ) {
                ArgusLog(LOG_INFO, "error creating additional-data meaning: %s.\n",
                             prelude_strerror(ret));
                return -1;
        }
        
        ret = prelude_string_set_ref(str, meaning);
        if ( ret < 0 ) {
                ArgusLog(LOG_INFO, "error setting byte string data meaning: %s.\n",
                             prelude_strerror(ret));
        }
                
        return -1;

  }

int PrintFlags(char *str, struct ArgusRecord *argus) {

  if ((argus->ahdr.type & ARGUS_FAR) &&
      (argus->argus_far.flow.ip_flow.ip_p == IPPROTO_TCP)) {

    struct ArgusTCPObject *tcp = NULL;
    unsigned int status;

    if (ArgusThisFarStatus & ARGUS_TCP_DSR_STATUS) {
      tcp = (struct ArgusTCPObject *)ArgusThisFarHdrs[ARGUS_TCP_DSR_INDEX];
      }

    if ((tcp != NULL) && ((status = tcp->state) != 0)) {
        if (status & ARGUS_SAW_SYN)         strcat (str, "s");
        if (status & ARGUS_SAW_SYN_SENT)    strcat (str, "S");
        if (status & ARGUS_CON_ESTABLISHED) strcat (str, "E");
        if (status & ARGUS_FIN)             strcat (str, "f");
        if (status & ARGUS_FIN_ACK)         strcat (str, "F");
        if (status & ARGUS_NORMAL_CLOSE)    strcat (str, "C");
        if (status & ARGUS_RESET)           strcat (str, "R");
        return(0);
        }
      }
  // string unchanged
  return(-1);
  }

int ip_data_to_alert (idmef_alert_t *alert, struct ArgusRecord *argus,
	unsigned char protocol) {
	int result;
	idmef_time_t *detect_time;
	idmef_classification_t *classification;
	prelude_string_t *str;
	idmef_reference_t *ref;
	idmef_node_t *node;
	idmef_source_t *source;
	idmef_target_t *target;
	idmef_address_t *address;
	idmef_service_t *service;
	idmef_impact_t *impact;
	idmef_assessment_t *assessment;
	idmef_impact_severity_t severity;
	

        struct ClassStruct *found;

	char flags[LINE_LEN];
	char *flagp;
	struct in_addr IPdst, IPsrc;
	char *IPdsts, *IPsrcs;
        static char MACsrcfield[64], MACdstfield[64];
	char *MACdsts=&MACdstfield[0];
        char *MACsrcs=&MACsrcfield[0];

	struct ArgusFlow *flow = &argus->argus_far.flow;
	arg_uint32 saddr = 0, daddr = 0;
	arg_uint16 sport = 0, dport = 0;


	// First init the global settings (eg detect time)

	/* Init DetectTime */
	result = idmef_time_new_from_timeval(&detect_time, &argus->argus_far.time.start);
        if ( result < 0 ) {
		return -1;
		}
	idmef_alert_set_detect_time(alert, detect_time);

        // now find a matching ClassStruct for the actual argus record
        found=find_class(mclass, argus);	

        if (found == NULL) {
          // this should not happen because raprelude has a
          // final classification that matches all traffic
          ArgusLog(LOG_INFO, "no classification found...");
          return -1;
          }

        if (found->action == ACTION_DROP) {
          return(ACTION_DROP);
          }

	/* Init Classification */
	result = idmef_alert_new_classification(alert, &classification);
	if ( result < 0 ) {
		return -1;
		}
        result = idmef_classification_new_text(classification, &str);
        if ( result < 0 ) {
                return result;
		}
        prelude_string_set_ref(str, found->classification);

	result = idmef_classification_new_reference(classification, &ref, -1);
        if ( result < 0 ) {
		return result;
		}

	result = idmef_reference_new_name(ref, &str);
        if ( result < 0 ) {
		return result;
		}

        idmef_reference_set_origin(ref, IDMEF_REFERENCE_ORIGIN_VENDOR_SPECIFIC);
        if ( idmef_reference_get_origin(ref) != IDMEF_REFERENCE_ORIGIN_VENDOR_SPECIFIC )
                prelude_string_set_ref(str, "whatever ID");
        else
                prelude_string_set_constant(str, "url");

        result = idmef_reference_new_url(ref, &str);
        if ( result < 0 ) {
                return result;
        	}

        prelude_string_sprintf(str, "%s", found->referenceurl);

	/* Source of Attack */
	result = idmef_alert_new_source(alert, &source, -1);
        if ( result < 0 ) {
                return result;
        	}
	 
	result = idmef_source_new_service(source, &service);
        if ( result < 0 ) {
                return result;
        	}

        if ((protocol == IPPROTO_TCP) || (protocol == IPPROTO_UDP)) {
	  sport = flow->ip_flow.sport;
	  idmef_service_set_port(service, sport);
	  }

	idmef_service_set_iana_protocol_number(service, protocol);

	// TODO: add additional information like IP4/IP6

	result = idmef_source_new_node(source, &node);
        if ( result < 0 ) {
                return result;
        	}

	result = idmef_node_new_address(node, &address, -1);
        if ( result < 0 ) {
                return result;
        	}

	result = idmef_address_new_address(address, &str);
        if ( result < 0 ) {
                return result;
        	}

	idmef_address_set_category(address, IDMEF_ADDRESS_CATEGORY_IPV4_ADDR);

	saddr = flow->ip_flow.ip_src;
	IPsrc.s_addr=htonl(saddr);
	IPsrcs=strdup(inet_ntoa(IPsrc));
	prelude_string_set_ref(str, IPsrcs);
        // FIX ME: the string IPsrcs must be freed at some time

        // add mac address if ACTION_MLOG or ACTION_STORE
        ArgusThisFarStatus = ArgusIndexRecord(argus, ArgusThisFarHdrs);
        if (((found->action == ACTION_MLOG) ||
             (found->action == ACTION_STORE)) &&
            (ArgusThisFarStatus & ARGUS_MAC_DSR_STATUS)) {
          struct ArgusMacStruct *macsrc = (struct ArgusMacStruct *) ArgusThisFarHdrs[ARGUS_MAC_DSR_INDEX];
          char *esrcString;
          esrcString = etheraddr_string ((u_char *)&macsrc->phys_union.ether.ethersrc);
          bzero(MACsrcs, 64);
          sprintf (MACsrcs, "%17.17s", esrcString);

          result = idmef_node_new_address(node, &address, -1);
          if ( result < 0 ) {
                return result;
        	}
	  result = idmef_address_new_address(address, &str);
          if ( result < 0 ) {
                return result;
        	}

	  prelude_string_set_ref(str, MACsrcs);
	  idmef_address_set_category(address, IDMEF_ADDRESS_CATEGORY_MAC);
          }


	/* Target of Attack */
	result = idmef_alert_new_target(alert, &target, -1);
        if ( result < 0 ) {
                return result;
        	}
	 
	result = idmef_target_new_service(target, &service);
        if ( result < 0 ) {
                return result;
        	}

        if ((protocol == IPPROTO_TCP) || (protocol == IPPROTO_UDP)) {
	  dport = flow->ip_flow.dport;
	  idmef_service_set_port(service, dport);
	  }

	protocol=flow->ip_flow.ip_p;
	idmef_service_set_iana_protocol_number(service, protocol);

	// TODO: add additional information like IP4/IP6

	result = idmef_target_new_node(target, &node);
        if ( result < 0 ) {
                return result;
        	}

	result = idmef_node_new_address(node, &address, -1);
        if ( result < 0 ) {
                return result;
        	}

	result = idmef_address_new_address(address, &str);
        if ( result < 0 ) {
                return result;
        	}

	daddr = flow->ip_flow.ip_dst;
	IPdst.s_addr=htonl(daddr);
	IPdsts=strdup(inet_ntoa(IPdst));
	prelude_string_set_ref(str, IPdsts);
	// FIX ME: the string IPdsts must be freed at some time

        // add mac address if ACTION_MLOG or ACTION_STORE
        ArgusThisFarStatus = ArgusIndexRecord(argus, ArgusThisFarHdrs);
        if (((found->action == ACTION_MLOG) ||
             (found->action == ACTION_STORE)) &&
            (ArgusThisFarStatus & ARGUS_MAC_DSR_STATUS)) {
          struct ArgusMacStruct *macdst = (struct ArgusMacStruct *) ArgusThisFarHdrs[ARGUS_MAC_DSR_INDEX];
          char *edstString;
          edstString = etheraddr_string ((u_char *)&macdst->phys_union.ether.etherdst);
          bzero(MACdsts, 64);
          sprintf (MACdsts, "%17.17s", edstString);

          result = idmef_node_new_address(node, &address, -1);
          if ( result < 0 ) {
                return result;
        	}
	  result = idmef_address_new_address(address, &str);
          if ( result < 0 ) {
                return result;
        	}
	  prelude_string_set_ref(str, MACdsts);
	  idmef_address_set_category(address, IDMEF_ADDRESS_CATEGORY_MAC);

          }

	// TODO: eventually add decoy field (or did this
	//	 vanish in the actual IDMEF draft?
	//target->decoy = unknown;
	

	// add some entries to additionalData
        add_int_additionalData(alert, "ip_proto", protocol);

	/* Set StartTime and EndTime */
	add_int_additionalData(alert, "starttime.sec",
		argus->argus_far.time.start.tv_sec);
	add_int_additionalData(alert, "starttime.usec",
		argus->argus_far.time.start.tv_usec);
	add_int_additionalData(alert, "lasttime.sec",
		argus->argus_far.time.last.tv_sec);
	add_int_additionalData(alert, "lasttime.usec",
		argus->argus_far.time.last.tv_usec);

        // Set byte and packet counts
        add_int_additionalData(alert, "packets.src2dst",
		argus->argus_far.src.count);
        add_int_additionalData(alert, "packets.dst2src",
		argus->argus_far.dst.count);
        add_int_additionalData(alert, "bytes.src2dst",
		argus->argus_far.src.bytes);
        add_int_additionalData(alert, "bytes.dst2src",
		argus->argus_far.dst.bytes);

        if (protocol == IPPROTO_TCP) {
          flagp=&flags[0];
          bzero(flagp, LINE_LEN);
          if (PrintFlags(flagp, argus)==0) {
	    add_string_additionalData(alert, "argusFlags", flagp);
            }
          }

        if (found->action == ACTION_STORE) {
          // store the complete argus record into additionalData
	  // first convert to network byte order (ArgusHtoN does
          // this only if necessary)
	  ArgusHtoN(argus);
	  add_bin_additionalData(alert, "argusrecord", (char *) argus, sizeof(struct ArgusRecord));
	  // and restore host byte order (if necessary)
	  ArgusNtoH(argus);
	  }

	/* Impact */

	result = idmef_alert_new_assessment(alert, &assessment);
        if ( result < 0 ) {
                return result;
        	}

	result = idmef_assessment_new_impact(assessment, &impact);
        if ( result < 0 ) {
                return result;
        	}

	severity = found->severity;
	idmef_impact_set_severity(impact, severity);

	// TODO: Maybe add completion, confidence rating & description
	//       to impact?

        return 0;
}

int send_ip_alert(struct ArgusRecord *argus, unsigned char proto) {

	idmef_alert_t *alert;
	idmef_message_t *message;
	idmef_time_t *create_time;
        int result;

	result = idmef_message_new(&message);
	if ( result < 0 ) {
	        ArgusLog(LOG_INFO, "send_ip_alert: idmef_message_new failed");
		goto err;
		}

	result = idmef_message_new_alert(message, &alert);
	if ( result < 0 ) {
	        ArgusLog(LOG_INFO, "send_ip_alert: idmef_message_new_alert failed");
		goto err;
		}

        result = ip_data_to_alert(alert, argus, proto);
        if ( result == ACTION_DROP ) {
                goto err;
                }
        else if (result < 0) {
	        ArgusLog(LOG_INFO, "send_ip_alert: ip_data_to_alert failed");
		goto err;
		}

	result = idmef_time_new_from_gettimeofday(&create_time);
        if ( result < 0 ) {
	        ArgusLog(LOG_INFO, "send_ip_alert: idmef_time_new failed");
		goto err;
		}
	idmef_alert_set_create_time(alert, create_time);
	idmef_alert_set_analyzer(alert, idmef_analyzer_ref(argus_analyzer), 0);

	prelude_client_send_idmef(pclient, message);

	/* Sending done. */
	idmef_message_destroy(message);
	if ( result < 0 ) {
		goto err;
		}
	return(0);
	
  err:
	idmef_message_destroy(message);
	return(-1);
}


void
RaProcessRecord (struct ArgusRecord *argus)
{
   struct ArgusRecordStore *store = NULL;
   int process = 0;
   unsigned char ip_proto;

   if (argus->ahdr.type & ARGUS_MAR)
      process++;

   else {
      switch (argus->ahdr.status & 0xFFFF) {
         case ETHERTYPE_IP:
            ip_proto=argus->argus_far.flow.ip_flow.ip_p;
            process++;
            send_ip_alert(argus, ip_proto);
            break;

         case ETHERTYPE_ARP:
         case ETHERTYPE_REVARP:
         default:
            break;
      }
   }


   if (process) {
         if ((store = RaNewArgusStore(argus)) != NULL) {
            if ((store->data[0] = RaNewArgusData(argus)) != NULL) {
               store->data[0]->status |= RA_MODIFIED;
               RaSendArgusRecord (store);
               RaDeleteArgusStore (store);

            } else
               ArgusLog (LOG_ERR, "RaProcessRecord: RaNewArgusData(0x%x) %s\n",
                                  argus, strerror(errno));
         } else
            ArgusLog (LOG_ERR, "RaProcessRecord: RaNewArgusStore(0x%x) %s\n",
                                  argus, strerror(errno));
   }

   if (Nflag > 0) {
      if (--Nflag == 0) {
         RaParseComplete(SIGINT);
      }
   }
}


int
RaSendArgusRecord(struct ArgusRecordStore *store)
{
   struct ArgusRecordData *data;
   struct ArgusRecord *argus = NULL;
   int retn = 0, i;

   for (i = 0; i < RaHistoTimeSeries; i++) {
      if ((data = store->data[i]) != NULL) {
         argus = data->argus;

         if (!aflag) {
#ifdef _LITTLE_ENDIAN
            ArgusHtoN(argus);
#endif
            ArgusGenerateCanonicalRecord (argus, RaThisCanon);
 
#ifdef _LITTLE_ENDIAN
            ArgusNtoH(argus);
#endif
            if ((retn = argus_filter (ArgusFilterCode.bf_insns, RaThisCanon)) == 0)
               return (retn);
         }

         if ((ArgusWfileList != NULL) && (!(ArgusListEmpty(ArgusWfileList)))) {
            struct ArgusWfileStruct *wfile = NULL, *start = NULL;
 
            if ((wfile = ArgusFrontList(ArgusWfileList)) != NULL) { 
               start = wfile; 
               do { 
                  if ((exceptfile == NULL) || strcmp(wfile->filename, exceptfile)) { 
#ifdef _LITTLE_ENDIAN
                     ArgusHtoN(argus); 
#endif
                     ArgusWriteNewLogfile (wfile, argus); 
 
#ifdef _LITTLE_ENDIAN
                     ArgusNtoH(argus); 
#endif
                  }   
                  ArgusPopFrontList(ArgusWfileList); 
                  ArgusPushBackList(ArgusWfileList, wfile); 
                  wfile = ArgusFrontList(ArgusWfileList); 
 
               } while (wfile != start); 
            }   

         } else {
            if (argus->ahdr.type & ARGUS_MAR)
               printf ("%s\n", get_man_string (argus));
                  
            else {

               if (Lflag && !(qflag)) {
                  if (RaLabel == NULL)
                     RaLabel = RaGenerateLabel(argus);

                  if (!(RaLabelCounter++ % Lflag)) {
                     printf ("%s\n", RaLabel);
                     fflush (stdout);
                  }

                  if (Lflag < 0)
                     Lflag = 0;
               }

               ArgusThisFarStatus = ArgusIndexRecord(argus, ArgusThisFarHdrs);

               switch (argus->ahdr.status & 0xFFFF) {
                  case ETHERTYPE_IP:
                     switch (argus->argus_far.flow.ip_flow.ip_p) {
                        case IPPROTO_TCP: { 
                           int srcportind = 0;
                           if (argus->argus_far.flow.ip_flow.sport == 20) {
                              if (RaPrintAlgorithms[3] == ArgusPrintDstPort) {
                                 RaPrintAlgorithms[3] = ArgusPrintSrcPort;
                                 srcportind++;
                              }
                           }
                           
                           printf ("%s", get_tcp_string (argus));

                           if (srcportind)
                              RaPrintAlgorithms[3] = ArgusPrintDstPort;
                           break;
                        }
         
                        case IPPROTO_ICMP:              
                           printf ("%s", get_icmp_string (argus));
                           break;
      
                        default:
                           printf ("%s", get_ip_string (argus));
                           break;
                     }
                     break;
         
                  case ETHERTYPE_ARP:
                  case ETHERTYPE_REVARP:
                     printf ("%s", get_arp_string (argus));
                     break;
         
                  default:
                     printf ("%s", get_nonip_string (argus));
                     break;
               }

               printf ("\n");
            }

            fflush (stdout);
         }
      }
   }

   store->qhdr.logtime = ArgusGlobalTime;

#ifdef ARGUSDEBUG
   ArgusDebug (4, "RaSendArgusRecord(0x%x) done.\n", store);
#endif

   return (retn);
}


// get the next word (or words included in quotes) from a buffer
char *get_next(char *thestring, char **next) {
  int count=0, scount=0, qcount=0;
  char quotes[LINE_LEN];
  char store=0; char *start=NULL, *tempstring=NULL;
  static char *temp=NULL;

  if (*next==NULL) {
    temp=strdup(thestring);
    *next=temp;
    }

  if (temp==NULL) {
    // not initialized
    return(NULL);
    }

  if (*next >= temp + strlen(temp)) {
    // no further arguments
    return(NULL);
    }

  start=*next;

  // initially quotes only contains a space
  quotes[0]=' '; quotes[1]=0; qcount=0;
  
  // skip leading spaces or tabs
  while ((scount < strlen(start)) &&
         ((start[scount] == ' ') || (start[scount] == '\t'))){
    scount++;
    }
  start=start+scount;

  count=0;
  while (count < strlen(start)) {
    if (start[count] == quotes[qcount]) {
      // matching quote or the final space found
      if ((start[count]=='\'') || (start[count] == '"')) {
        // closing quote so remove from qcount
        quotes[qcount]=0; qcount--;
        }
      else if ((start[count]==' ') && (qcount==0)) {
        // final space
        *next=start+scount+count;
        start[count]=0;
        tempstring=strdup(start); start[count]=' ';
        // ArgusLog(LOG_INFO, "get_next: word %s", tempstring);
        return(tempstring);
        }
      }
    else if ((start[count]=='\'') || (start[count]=='\"')) {
      // opening quote, so store in quotes
      qcount++; quotes[qcount]=start[count]; quotes[qcount+1]=0;
      if (qcount >= (LINE_LEN - 1)) {
        ArgusLog(LOG_INFO, "get_next: too many nested quotes in line");
        return(NULL);
        }
      }
    count++;
    } // end of while

  *next=start + strlen(start);
  return(strdup(start));
  }

int parse_class (struct ClassStruct *classif, char *buf) {

  char *act_word, *nxt_word=NULL;
  int state=0, error = -1, count;
  char sbuffer[LINE_LEN];
  char *sbuf = &sbuffer[0];
  struct protoent *protocol_by_name;

  // ArgusLog(LOG_INFO, "parse_class: parsing line %d", classif->lineno);
  
  act_word=get_next(buf, &nxt_word);
  if (act_word == NULL) {
    // empty line somehow...
    return PARSERR_NOSPLIT;
    }

  while ((act_word != NULL) && (*act_word != '\n') && (error==-1)) {
    state=-1;
    for (count=0; count < CONFIG_WORDS_COUNT; count++) {
      if ((strcmp(act_word, config_words[count]))==0) {
        state=count;
        }
      }
    if (state < 0) {
      // no allowed word from configuration
      return(PARSERR_INVWORD);
      }
    switch (state) {
      case STATE_PROTO:
        act_word=get_next(buf, &nxt_word);
        if (strcmp(act_word, "any")==0) {
          classif->proto = (arg_uint16) PROTO_ANY;
          act_word=get_next(buf, &nxt_word);
          break;
          }
        else {
          protocol_by_name = getprotobyname(act_word);
          if (protocol_by_name == NULL) {
            return(PARSERR_INVPROTO);
            }
          classif->proto = (arg_uint16) protocol_by_name->p_proto;
          act_word=get_next(buf, &nxt_word);
          break;
          }
      case STATE_SADDR:
        act_word=get_next(buf, &nxt_word);
        if (strcmp(act_word, "any")==0) {
          // nothing necessary (classif already initialized to "any")
          act_word=get_next(buf, &nxt_word);
          }
        else {
          classif->src.addr = ntohl(inet_addr(get_next(buf, &nxt_word)));
          classif->src.mask = ntohl(inet_addr(get_next(buf, &nxt_word)));
          act_word=get_next(buf, &nxt_word);
          }
        break;
      case STATE_SPORT:
        classif->src_port_action=NO_PORT;
        act_word=get_next(buf, &nxt_word); 
        if ((strncmp(act_word, "eq", 2))==0) {
          classif->src_port_action=PORT_EQ;
          if (!(classif->src_port_low = (arg_uint16) atoi (get_next(buf, &nxt_word)))) {
            return(PARSERR_SPORT);
            }
          act_word=get_next(buf, &nxt_word);
          }
        else if ((strncmp(act_word, "gt", 2))==0) {
          classif->src_port_action=PORT_GT;
          if (!(classif->src_port_low = (arg_uint16) atoi (get_next(buf, &nxt_word)))) {
            return(PARSERR_SPORT);
            }
          act_word=get_next(buf, &nxt_word);
          }
        else if ((strncmp(act_word, "lt", 2))==0) {
          classif->src_port_action=PORT_LT;
          if (!(classif->src_port_low = (arg_uint16) atoi (get_next(buf, &nxt_word)))) {
            return(PARSERR_SPORT);
            }
          act_word=get_next(buf, &nxt_word);
          }
        else if ((strncmp(act_word, "ne", 2))==0) {
          classif->src_port_action=PORT_NE;
          if (!(classif->src_port_low = (arg_uint16) atoi (get_next(buf, &nxt_word)))) {
            return(PARSERR_SPORT);
            }
          act_word=get_next(buf, &nxt_word);
          }
        else if ((strncmp(act_word, "range", 5))==0) {
          classif->src_port_action=PORT_RANGE;
          if (!(classif->src_port_low = (arg_uint16) atoi (get_next(buf, &nxt_word)))) {
            return(PARSERR_SPORT);
            }
          if (!(classif->src_port_hi = (arg_uint16) atoi (get_next(buf, &nxt_word)))) {
            return(PARSERR_SPORT);
            }
          act_word=get_next(buf, &nxt_word);
          }
        else {
          return(PARSERR_SPORTOP);
          }
          break;
      case STATE_DADDR:
        act_word = get_next(buf, &nxt_word);
        if (strcmp(act_word, "any")==0) {
          // nothing necessary (classif already initialized to "any")
          act_word=get_next(buf, &nxt_word);
          if (nxt_word == NULL) { break; }
          }
        else {
          classif->dst.addr = ntohl(inet_addr(get_next(buf, &nxt_word)));
          classif->dst.mask = ntohl(inet_addr(get_next(buf, &nxt_word)));
          act_word=get_next(buf, &nxt_word);
          }
          break;
        // and now parse the port number specified
      case STATE_DPORT:
        classif->dst_port_action=NO_PORT;
        act_word=get_next(buf, &nxt_word);
        if ((strncmp(act_word, "eq", 2))==0) {
          classif->dst_port_action=PORT_EQ;
          if (!(classif->dst_port_low = (arg_uint16) atoi (get_next(buf, &nxt_word)))) {
            return(PARSERR_DPORT);
            }
          act_word=get_next(buf, &nxt_word);
          }
        else if ((strncmp(act_word, "gt", 2))==0) {
          classif->dst_port_action=PORT_GT;
          if (!(classif->dst_port_low = (arg_uint16) atoi (get_next(buf, &nxt_word)))) {
            return(PARSERR_DPORT);
            }
          act_word=get_next(buf, &nxt_word);
          }
        else if ((strncmp(act_word, "lt", 2))==0) {
          classif->dst_port_action=PORT_LT;
          if (!(classif->dst_port_low = (arg_uint16) atoi (get_next(buf, &nxt_word)))) {
            return(PARSERR_DPORT);
            }
          act_word=get_next(buf, &nxt_word);
          }
        else if ((strncmp(act_word, "ne", 2))==0) {
          classif->dst_port_action=PORT_NE;
          if (!(classif->dst_port_low = (arg_uint16) atoi (get_next(buf, &nxt_word)))) {
            return(PARSERR_DPORT);
            }
          act_word=get_next(buf, &nxt_word);
          }
        else if ((strncmp(act_word, "range", 5))==0) {
          classif->dst_port_action=PORT_RANGE;
          if (!(classif->dst_port_low = (arg_uint16) atoi (get_next(buf, &nxt_word)))) {
            return(PARSERR_DPORT);
            }
          if (!(classif->dst_port_hi = (arg_uint16) atoi (get_next(buf, &nxt_word)))) {
            return(PARSERR_DPORT);
            }
          act_word=get_next(buf, &nxt_word);
          }
        else {
          return(PARSERR_DPORTOP);
          }
        break;
      // or maybe an icmp type was specified?
      case STATE_ITYPE:
        classif->itype=NO_ITYPE;
        if (!(classif->itype= (arg_uint16) atoi (get_next(buf, &nxt_word)))) {
            return(PARSERR_ITYPE);
            }
          act_word=get_next(buf, &nxt_word);
          if (nxt_word == NULL) { break; }
        break;
      case STATE_CLASS:
        classif->classification=NULL;
        act_word=get_next(buf, &nxt_word);
        if (*act_word == '\"') {
          // classification in double qoutes, so get the rest
          strcpy(sbuf, act_word+1); sbuf[strlen(sbuf)-1]=0;
          }
        else if (*act_word == '\'') {
          // classification in single qoutes, so get the rest
          strcpy(sbuf, act_word+1); sbuf[strlen(sbuf)-1]=0;
          }
        else {
          // no quotes, so it's only a single word
          strcpy(sbuf, act_word);
          }
        classif->classification = strdup(sbuf);
        // ArgusLog(LOG_INFO, "  Classification: %s", classif->classification);
        act_word=get_next(buf, &nxt_word);
        break;
      case STATE_URL:
        classif->referenceurl=NULL;
        act_word=get_next(buf, &nxt_word);
        if (*act_word == '\"') {
          // URL in double qoutes, so get the rest
          strcpy(sbuf, act_word+1); sbuf[strlen(sbuf)-1]=0;
          }
        else if (*act_word == '\'') {
          // URL in single qoutes, so get the rest
          strcpy(sbuf, act_word+1); sbuf[strlen(sbuf)-1]=0;
          }
        else {
          // no quotes, so it's only a single word
          strcpy(sbuf, act_word);
          }
        classif->referenceurl = strdup(sbuf);
        act_word=get_next(buf, &nxt_word);
        // ArgusLog(LOG_INFO, "  URL: %s", classif->referenceurl);
        break;
      case STATE_SEVERITY:
        act_word=get_next(buf, &nxt_word);
        classif->severity=Parse_idmef_impact_severity(act_word);
        if (classif->severity == -1) {
          return(PARSERR_SEVER);
          }
        act_word = get_next(buf, &nxt_word);
        break;
      case STATE_ACTION:
        act_word=get_next(buf, &nxt_word);
        if (strncmp(act_word, "log", 3)==0) {
          classif->action=ACTION_LOG; }
        else if (strncmp(act_word, "mlog", 4)==0) {
          classif->action=ACTION_MLOG;
          }
        else if (strncmp(act_word, "drop", 4)==0) {
          classif->action=ACTION_DROP;
          }
        else if (strncmp(act_word, "store", 4)==0) {
          classif->action=ACTION_STORE;
          }
        else {
          return(PARSERR_ACTION);
          }
        act_word = get_next(buf, &nxt_word);
        break;
      default:
        // this should never happen as unknown key words are already
        // processed before the switch statement
        return(PARSERR_INVWORD);
        break;
        } // end of switch
      state++; 
    } // end of while
  return 0;

  }

int init_class(struct ClassStruct *classif) {
  classif->src.addr=ntohl(inet_addr("0.0.0.0"));
  classif->src.mask=ntohl(inet_addr("255.255.255.255"));
  classif->src.operator=0; // this is not used by raprelude
  classif->dst.addr=ntohl(inet_addr("0.0.0.0"));
  classif->dst.mask=ntohl(inet_addr("255.255.255.255"));
  classif->dst.operator=0; // this is not used by raprelude
  classif->proto=PROTO_ANY;
  classif->src_port_low=0;
  classif->src_port_hi=MAXPORT;
  classif->dst_port_low=0;
  classif->dst_port_hi=MAXPORT;
  classif->src_port_action=NO_PORT;
  classif->dst_port_action=NO_PORT;
  classif->itype=ITYPE_ANY;
  classif->severity=IDMEF_IMPACT_SEVERITY_INFO;
  classif->classification=NULL;
  classif->referenceurl=NULL;
  classif->action=ACTION_DROP;
  }

int read_classfile(struct ClassStruct **mclass, char *filename) {
  FILE *classfile;
  struct ClassStruct *actclass, *lastclass = NULL;
  int linenum=0;
  char buffer[LINE_LEN];
  int result;

  ArgusLog(LOG_INFO, "read_classfile called with file %s", filename);

  classfile=fopen(filename, "r");
  if (classfile == NULL) {
    ArgusLog(LOG_DEBUG, "no class configuration, using default.");
    return(-1);
    }

  while (fgets(buffer, LINE_LEN, classfile)) {
    linenum++;
    if ((*buffer != '#') && (*buffer != '\n')) {
      if ((actclass = (struct ClassStruct *) calloc (1, sizeof(struct ClassStruct))) != NULL) {
        init_class(actclass);
        actclass->line = strdup(buffer);
        actclass->lineno = linenum;
        result = parse_class(actclass, buffer);
        if (result != 0) {
          ArgusLog(LOG_ERR, "parse error in file %s, line %d: %s",
              filename, linenum, parse_error[result]);
          } // end of if parseerror
        if (lastclass) {
          lastclass->nxt = actclass;
          actclass->prv = lastclass;
          lastclass = actclass;
          }
        else {
          // this is for the first entry, so store the first
          // allocated element in global variable mclass
          *mclass = lastclass = actclass;
          }
        } // end of if calloc
      } // end of if commentline or emptyline
    } // end of while

  fclose(classfile);

  // and finally add a default entry that matches everything
  // just to make sure we "know" what to do with everything
  // not specified in the config file
  if ((actclass = (struct ClassStruct *) calloc (1, sizeof(struct ClassStruct))) != NULL) {
    init_class(actclass);
    // FIX ME: add some more things that match everything?
    actclass->classification=strdup(IDMEF_CLASSIF_NAME);
    actclass->referenceurl=strdup(IDMEF_CLASSIF_URL);
    actclass->action=ACTION_LOG;
    actclass->src_port_low=0;
    actclass->src_port_hi=MAXPORT;
    actclass->src_port_action=PORT_RANGE;
    actclass->dst_port_low=0;
    actclass->dst_port_hi=MAXPORT;
    actclass->dst_port_action=PORT_RANGE;
    lastclass->nxt = actclass;
    actclass->prv = lastclass;
    }
  }


int IDS_init_connection(char *profile_name) {

  int result, noclass;
  prelude_string_t *string;
  prelude_client_profile_t *profile;

  ArgusLog(LOG_INFO, "IDS_init_connection called. Profile %s",
  	profile_name);

  result=prelude_init(NULL, NULL);

  if (result<0) {
    ArgusLog(LOG_ERR, "prelude_init failed with result %i",
    		result);
    }
    
  result = prelude_client_new(&pclient, profile_name ? profile_name : DEFAULT_ANALYZER_NAME);
  if ( result < 0 ) {
    ArgusLog(LOG_ERR, "unable to create a prelude client object: %i.",
                result);
    }

  // parse config file for IDMEF alert classification
  profile=prelude_client_get_profile(pclient);
  prelude_client_profile_get_profile_dirname(profile,
      classfile, LINE_LEN-strlen(CLASS_CFG_FILE)-1-1);

  strncat(classfile, "/", LINE_LEN - 1);
  strncat(classfile, CLASS_CFG_FILE, LINE_LEN - 1);

  if (read_classfile(&mclass, classfile) != 0) {
    noclass=FALSE;
    }
  else {
    noclass=TRUE;
    }

  // setup analyzer.

  argus_analyzer = prelude_client_get_analyzer(pclient);

  result = idmef_analyzer_new_model(argus_analyzer, &string);
  if ( result < 0 ) {
    ArgusLog(LOG_ERR, "unable to set analyzermodel: %i", result);
     }
  prelude_string_set_constant(string, IDMEF_ANALYZER_MODEL);

  result = idmef_analyzer_new_class(argus_analyzer, &string);
  if ( result < 0 ) {
    ArgusLog(LOG_ERR, "unable to set analyzer class: %i", result);
     }
  prelude_string_set_constant(string, IDMEF_ANALYZER_CLASS);

  result = idmef_analyzer_new_manufacturer(argus_analyzer, &string);
  if ( result < 0 ) {
    ArgusLog(LOG_ERR, "unable to set analyzer manufacturer: %i", result);
     }
  prelude_string_set_constant(string, IDMEF_ANALYZER_MANUFACTURER);

  result = idmef_analyzer_new_version(argus_analyzer, &string);
  if ( result < 0 ) {
    ArgusLog(LOG_ERR, "unable to set analyzer version: %i", result);
     }
  prelude_string_set_constant(string, IDMEF_ANALYZER_VERSION);

  result = prelude_client_start(pclient);
  if ( result < 0 ) {
    if ( prelude_client_is_setup_needed(result)) {
      prelude_client_print_setup_error(pclient);
      }
    ArgusLog(LOG_ERR, "unable to start prelude client: %i", result);
    }

  return(0);
}

void
ArgusClientInit ()
{

   int result=0;

   if (!(RaInitialized)) {
      RaInitialized++;

      result=IDS_init_connection(getenv("RAPRELUDE_PROFILE"));

   }

}
